.version-picker {
    display: flex;
    align-items: center;
}

.version-picker .dropdown {
    width: 130px;
    max-height: 29px;
    margin-left: 10px;
    display: inline-block;
    border-radius: 4px;
    border: 1px solid var(--theme-popup-border);
    position: relative;
    font-size: 13px;
    color: var(--fg);
    height: 100%;
    text-align: left;
}
.version-picker .dropdown .select {
    cursor: pointer;
    display: block;
    padding: 5px 2px 5px 15px;
}
.version-picker .dropdown .select > i {
    font-size: 10px;
    color: var(--fg);
    cursor: pointer;
    float: right;
    line-height: 20px !important;
}
.version-picker .dropdown:hover {
    border: 1px solid var(--theme-popup-border);
}
.version-picker .dropdown:active {
    background-color: var(--theme-popup-bg);
}
.version-picker .dropdown.active:hover,
.version-picker .dropdown.active {
    border: 1px solid var(--theme-popup-border);
    border-radius: 2px 2px 0 0;
    background-color: var(--theme-popup-bg);
}
.version-picker .dropdown.active .select > i {
    transform: rotate(-180deg);
}
.version-picker .dropdown .dropdown-menu {
    position: absolute;
    background-color: var(--theme-popup-bg);
    width: 100%;
    left: -1px;
    right: 1px;
    margin-top: 1px;
    border: 1px solid var(--theme-popup-border);
    border-radius: 0 0 4px 4px;
    overflow: hidden;
    display: none;
    max-height: 300px;
    overflow-y: auto;
    z-index: 9;
}
.version-picker .dropdown .dropdown-menu li {
    font-size: 12px;
    padding: 6px 20px;
    cursor: pointer;
} 
.version-picker .dropdown .dropdown-menu {
    padding: 0;
    list-style: none;
}
.version-picker .dropdown .dropdown-menu li:hover {
    background-color: var(--theme-hover);
}
.version-picker .dropdown .dropdown-menu li.active::before {
    display: inline-block;
    content: "✓";
    margin-inline-start: -14px;
    width: 14px;
}