{% extends "_base.html" %}
{% block title %}Authentication{% endblock %}

{% block header %}
<style type="text/css">
  #registrationForm input {
    display: block;
    margin: auto;
  }
</style>
{% endblock %}

{% block body %}
<form id="registrationForm" method="post" action="{{ myurl }}">
    <div>
        {% if error is defined %}
            <p class="error"><strong>Error: {{ error }}</strong></p>
        {% endif %}
        <p>
            Please click the button below if you agree to the
            <a href="{{ terms_url }}">privacy policy of this homeserver.</a>
        </p>
        <input type="hidden" name="session" value="{{ session }}" />
        <input type="submit" value="Agree" />
    </div>
</form>
{% endblock %}
